﻿using System;
using System.Collections.Generic;
using System.Text;

namespace pl0
{
    /// <summary>
    /// Kompilace bloku
    /// </summary>
    class Block
    {
	    private static LexicalA lexicalA;
	    
	    private static SymbolT symbol;
	    
	    private static CodeGen code;
	    
	    private static BugBox bug = new BugBox();

	    private const int LEVEL_MAX = 3;

        private int dataIndex = 3;  
        
        private int level;
       
        private int tx;
        
        private SetSym fsys;
        
        private int codeOffsetBegin;

        private BegSys bs = new BegSys();

        private void error(int ErrorType)
        {
            lexicalA.writeError(ErrorType);
            bug.error(ErrorType);
        }

        /// <summary>
        /// Nastaveni BugBoxu pro blok 
        /// </summary>
        /// <param name="bugbox">nastavovany BugBox</param>
        public void setBugBox(BugBox bugbox)
        {
            bug = bugbox;
        }
        /// <summary>
        /// Vraceni BugBox pouzivaneho blokem 
        /// </summary>
        /// <returns>pouzivany blok</returns>
        public BugBox getBugBox()
        {
            return bug;
        }

        /// <summary>
        /// Konstuktoru bloku 
        /// </summary>
        /// <param name="plevel">uroven zanoreni bloku</param>
        /// <param name="ptx">aktualni index na konec tabulky symbolu</param>
        /// <param name="pfsys">akceptovatelne znaky</param>
        public Block(int plevel, int ptx, SetSym pfsys)
        {
            fsys = pfsys;
            tx = ptx;
            level = plevel;
        }
        
        /// <summary>
        /// Spusteni prekladu nejvyssiho bloku v hierarchii
        /// </summary>
        /// <param name="lex">pouzity lexikalni analyzator</param>
        /// <param name="sym">akceptovatelna mnozina symbolu</param>
        /// <param name="gen">pouzity generator kodu</param>
        public void run(LexicalA lex, SymbolT sym, CodeGen gen)
        {
            lexicalA = lex;
            symbol = sym;
            code = gen;
            block();
        }

        private void constDeclaration()
        {
            if (lexicalA.getCurrentSymbol() == SetSym.IDENT)
            {
                lexicalA.getNextSymbol();
                if ((lexicalA.getCurrentSymbol() == SetSym.EQL) ||
                     (lexicalA.getCurrentSymbol() == SetSym.BECOMES))
                {
                    if (lexicalA.getCurrentSymbol() == SetSym.BECOMES)
                        error(BugBox.BAD_BECOMES);
                    lexicalA.getNextSymbol();
                    if (lexicalA.getCurrentSymbol() == SetSym.NUMBER)
                    {
                        if (lexicalA.isLastNumberTooBig())                   
                            error(BugBox.TOO_BIG_NUMBER);                    
                       
                        symbol.enterConstant(lexicalA.getCurrentIdentifier(),
                                             level,                          
                                             lexicalA.getCurrentNumber());   
                       
                        lexicalA.getNextSymbol();
                    }
                    else
                    {
                        error(BugBox.NUM_REQ);
                    }
                }
                else
                {
                    error(BugBox.EQUALS_REQ);
                }
            }
            else
            {
                error(BugBox.ID_REQ);
            }
        }

        private void varDeclaration(int level)
        {
            if (lexicalA.getCurrentSymbol() == SetSym.IDENT)
            {
                symbol.enterVariable(lexicalA.getCurrentIdentifier(),
                                     dataIndex, level);
                lexicalA.getNextSymbol();
                dataIndex++;
            }
            else
            {
                error(BugBox.ID_REQ);
            }
        }

       /// <summary>
       /// Zpracovani bloku
       /// </summary>
        private void block()
        {
            int tx0;
            SetSym pom = new SetSym();
            
            Block subblock;

            tx0 = tx;
            symbol.setAddr(tx, code.getAddressCounter());
            code.gen(Fct.JMP, 0, 0);
            if (level > LEVEL_MAX)
                error(BugBox.TOO_BIG_LEVEL);
            do
            {
                if (lexicalA.getCurrentSymbol() == SetSym.CONSTSYM)
                {
                    
                    lexicalA.getNextSymbol();
                    do
                    {
                        constDeclaration();
                        while (lexicalA.getCurrentSymbol() == SetSym.COMMA)
                        {
                            lexicalA.getNextSymbol();
                            constDeclaration();
                        }
                        if (lexicalA.getCurrentSymbol() == SetSym.SEMICOLON)
                        {
                            lexicalA.getNextSymbol();
                        }
                        else
                        {
                            error(BugBox.SEMI_COMO_REQ);
                        }
                    } while (lexicalA.getCurrentSymbol() == SetSym.IDENT);
                }
                if (lexicalA.getCurrentSymbol() == SetSym.VARSYM)
                {
                    
                    lexicalA.getNextSymbol();
                    varDeclaration(level);
                    while (lexicalA.getCurrentSymbol() == SetSym.COMMA)
                    {
                        lexicalA.getNextSymbol();
                        varDeclaration(level);
                    }
                    if (lexicalA.getCurrentSymbol() == SetSym.SEMICOLON)
                    {
                        lexicalA.getNextSymbol();
                    }
                    else
                    {
                        error(BugBox.SEMI_COMO_REQ);
                    }
                }
                while (lexicalA.getCurrentSymbol() == SetSym.PROCSYM)
                {
                   
                    lexicalA.getNextSymbol();
                    if (lexicalA.getCurrentSymbol() == SetSym.IDENT)
                    {
                        symbol.enterProcedure(lexicalA.getCurrentIdentifier(),
                                              level);
                        lexicalA.getNextSymbol();
                    }
                    else
                    {
                        error(BugBox.ID_REQ);
                    }
                    if (lexicalA.getCurrentSymbol() == SetSym.SEMICOLON)
                    {
                        lexicalA.getNextSymbol();
                    }
                    else
                    {
                        error(BugBox.SEMI_COMO_REQ);
                    }
                    pom = pom.or(fsys);
                    pom.set(SetSym.SEMICOLON);
                    subblock = new Block(level + 1, symbol.getTableIndex(), pom);
                    subblock.block();
                    if (lexicalA.getCurrentSymbol() == SetSym.SEMICOLON)
                    {
                        lexicalA.getNextSymbol();
                        pom = new SetSym();
                        
                        pom = pom.or(bs.statBegSys);
                        pom.set(SetSym.IDENT);
                        pom.set(SetSym.PROCSYM);
                        test(pom, fsys, 6);
                    }
                    else
                    {
                        error(BugBox.SEMI_COMO_REQ);
                    }
                }
                pom = new SetSym();
                
                pom = pom.or(bs.statBegSys);
                pom.set(SetSym.IDENT);
                
                test(pom, bs.declBegSys, 7);
            } while (bs.declBegSys.get(lexicalA.getCurrentSymbol()));
            code.setAddressAt(((Locatable)symbol.getIdentifier(tx0)).getAddress(),
                              code.getAddressCounter());
            symbol.setAddr(tx0, code.getAddressCounter());
            codeOffsetBegin = code.getAddressCounter();
            code.gen(Fct.ING, 0, dataIndex);
            pom = new SetSym();
            pom = pom.or(fsys);
            pom.set(SetSym.SEMICOLON);
            pom.set(SetSym.ENDSYM);
            statement(pom, tx, level);
            code.gen(Fct.RET, 0, 0);
            test(fsys, pom, 8);
            code.listCode(codeOffsetBegin, code.getAddressCounter());
        }

        /// <summary>
        /// Generovani kodu pro ulozeni hodnoty jednoho faktoru na vrcholu zasobniku,
        /// </summary>
        /// <param name="fsys">mnozina dostupnych symbolu(tj.faktor muze byt pouze
        /// promenna, kontanta,cislo nebo levela zavorka)</param>
        /// <param name="tx">index na konec tabulky symbolu</param>
        /// <param name="level">uroven faktoru</param>
        private void factor(SetSym fsys, int tx, int level) 
        {
		    int i;
		    // pomocna promenna pouzivana pro vytvareni nove mnoziny symbolu
		    SetSym pom;
		    // aktualni identifikator
		    Identifier ident;

            test(bs.facBegSys, fsys, 24);
            while (bs.facBegSys.get(lexicalA.getCurrentSymbol()))
            {
			    if (lexicalA.getCurrentSymbol() == SetSym.IDENT) {       
				    i = symbol.position(lexicalA.getCurrentIdentifier(),     
				                        level);                              
				    if (i == 0) {
					    error(BugBox.UNDECLARED_ID);
				    } else {
					    ident = symbol.getIdentifier(i);
					    if ((ident.GetType()).IsInstanceOfType(new Constant(null,0,0))) {
						    code.gen(Fct.LIT, 0, ((Constant)ident).getValue());
					    } else if ((ident.GetType()).IsInstanceOfType(new Variable(null,0,0))) {
						    code.gen(Fct.LOD,
						             level-((Variable)ident).getLevel(),
						             ((Variable) ident).getAddress());
					    } else {
						    error(BugBox.INVALID_PROC_USE);
					    }
				    }
				    lexicalA.getNextSymbol();
			    } else if (lexicalA.getCurrentSymbol() == SetSym.NUMBER) {
				   
				    if (lexicalA.isLastNumberTooBig())                       
					    error(BugBox.TOO_BIG_NUMBER);                        
				    code.gen(Fct.LIT, 0, lexicalA.getCurrentNumber());
				    lexicalA.getNextSymbol();
			    } else {
				    if (lexicalA.getCurrentSymbol() == SetSym.LPAREN) {
					    lexicalA.getNextSymbol();                            
					    pom = new SetSym();
					    pom = pom.or(fsys);
					    pom.set(SetSym.RPAREN);
					    expression(pom,tx,level);
					    if (lexicalA.getCurrentSymbol() == SetSym.RPAREN) {
						    lexicalA.getNextSymbol();
					    } else {
						    error(BugBox.RPARENT_REQ);
					    }
				    }
			    }
			    pom = new SetSym();
			    pom.set(SetSym.LPAREN);
			    test(fsys, pom, BugBox.INVALID_FACTOR_END);
		    }  // while
	    }
        
        /// <summary>
        /// Generovani kodu pro vyhodnoceni jednoho termu
        /// </summary>
        /// <param name="fsys">mnozina dostupnych symbolu ve termu</param>
        /// <param name="tx">ukazovatko na konec tabulky symbolu</param>
        /// <param name="level">uroven termu</param>
        private void term(SetSym fsys, int tx, int level)
        {
            // operator termu
            int mulop;
            // promenna obsahujici aktualni symbol
            int sym;

            SetSym pom = new SetSym();
            pom = pom.or(fsys);
            pom.set(SetSym.TIMES);
            pom.set(SetSym.SLASH);
            pom.set(SetSym.MODULO);
            factor(pom, tx, level);

            sym = lexicalA.getCurrentSymbol();
            while ((sym == SetSym.TIMES) || (sym == SetSym.SLASH) ||
                    (sym == SetSym.MODULO))
            {
                mulop = sym;
                lexicalA.getNextSymbol();
                pom = pom.or(fsys);
                pom.set(SetSym.TIMES);
                pom.set(SetSym.SLASH);
                pom.set(SetSym.MODULO);
                factor(pom, tx, level);
                switch (mulop)
                {
                    case SetSym.TIMES:
                        code.gen(Fct.OPR, 0, OOperation.MUL);
                        break;
                    case SetSym.SLASH:
                        code.gen(Fct.OPR, 0, OOperation.DI);
                        break;
                    case SetSym.MODULO:
                        code.gen(Fct.OPR, 0, OOperation.MOD);
                        break;
                }
                sym = lexicalA.getCurrentSymbol();
            }
        }
        
        /// <summary>
        /// Generovani kodu pro vyhodnoceni aritm. vyrazu.
        /// </summary>
        /// <param name="fsys">mnozina dostupnych symbolu ve vyrazu</param>
        /// <param name="tx">ukazovatko na konec tabulky symbolu</param>
        /// <param name="level">uroven vyrazu</param>
        private void expression(SetSym fsys, int tx, int level)
        {
            // pouzity relacni operator
            int addop;
            // promenna obsahujici aktualni symbol
            int sym;
            // pomocna promenna obsahujici kod aktualniho symbolu		
            SetSym pom = new SetSym();

            sym = lexicalA.getCurrentSymbol();
            if ((sym == SetSym.PLUS) || (sym == SetSym.MINUS))
            {
                addop = sym;
                lexicalA.getNextSymbol();
                pom = new SetSym();
                pom = pom.or(fsys);
                pom.set(SetSym.PLUS);
                pom.set(SetSym.MINUS);
                term(pom, tx, level);
                if (addop == SetSym.MINUS)
                    code.gen(Fct.OPR, 0, OOperation.NEG);
            }
            else
            {
                pom = new SetSym();
                pom = pom.or(fsys);
                pom.set(SetSym.PLUS);
                pom.set(SetSym.MINUS);
                term(pom, tx, level);
            }
            sym = lexicalA.getCurrentSymbol();
            while ((sym == SetSym.PLUS) || (sym == SetSym.MINUS))
            {
                addop = sym;
                lexicalA.getNextSymbol();
                pom = new SetSym();
                pom = pom.or(fsys);
                pom.set(SetSym.PLUS);
                pom.set(SetSym.MINUS);
                term(pom, tx, level);
                if (addop == SetSym.PLUS)
                {
                    code.gen(Fct.OPR, 0, OOperation.ADD);
                }
                else
                {
                    code.gen(Fct.OPR, 0, OOperation.SUB);
                }
                sym = lexicalA.getCurrentSymbol();
            }
        }
        
        /// <summary>
        /// Generovani kodu pro vyhodnoceni logickeho vyrazu
        /// </summary>
        /// <param name="fsys">dostupne symoboly v vyrazu</param>
        /// <param name="tx">ukazovatko na konec tabulky symbolu</param>
        /// <param name="level">uroven, ve ktere je prelozeny vyraz</param>
        private void condition(SetSym fsys, int tx, int level)
        {
            // pouzita relace
            int relop;
            // pomocna promenna	pro vytvareni nove mnoziny symbolu
            SetSym pom;
            // pomocna promenna obsahujici kod aktualniho symbolu
            int sym;

            if (lexicalA.getCurrentSymbol() == SetSym.ODDSYM)
            {
                lexicalA.getNextSymbol();
                expression(fsys, tx, level);
                code.gen(Fct.OPR, 0, OOperation.ODD);
            }
            else
            {
                pom = new SetSym();
                pom = pom.or(fsys);
                pom.set(SetSym.EQL); pom.set(SetSym.NEQ);
                pom.set(SetSym.LSS); pom.set(SetSym.GTR);
                pom.set(SetSym.LEQ); pom.set(SetSym.GEQ);
                expression(pom, tx, level);
                sym = lexicalA.getCurrentSymbol();
                if ((sym != SetSym.EQL) && (sym != SetSym.NEQ) &&
                     (sym != SetSym.LSS) && (sym != SetSym.GTR) &&
                     (sym != SetSym.LEQ) && (sym != SetSym.GEQ))
                {
                    error(BugBox.RPARENT_REQ);
                }
                else
                {
                    relop = sym;
                    lexicalA.getNextSymbol();
                    expression(fsys, tx, level);
                    switch (relop)
                    {
                        case SetSym.EQL:
                            code.gen(Fct.OPR, 0, OOperation.EQ);
                            break;
                        case SetSym.NEQ:
                            code.gen(Fct.OPR, 0, OOperation.NE);
                            break;
                        case SetSym.LSS:
                            code.gen(Fct.OPR, 0, OOperation.LT);
                            break;
                        case SetSym.GEQ:
                            code.gen(Fct.OPR, 0, OOperation.GE);
                            break;
                        case SetSym.GTR:
                            code.gen(Fct.OPR, 0, OOperation.GT);
                            break;
                        case SetSym.LEQ:
                            code.gen(Fct.OPR, 0, OOperation.LE);
                            break;
                    }
                }
            }
        }

        /// <summary>
        /// Generovani kodu pro prikaz
        /// </summary>
        /// <param name="fsys">dostupne symboly v statmentu</param>
        /// <param name="tx">ukazovatko na konec tabulky symbolu</param>
        /// <param name="level">uroven, ve ktere je prelozeny statement</param>
        private void statement(SetSym fsys, int tx, int level) 
        {
		    int i, cx1, cx2;
		    // pomocna promenna obsahujici kod aktualniho symbolu
		    int sym;

		    SetSym pom = new SetSym();
		    sym = lexicalA.getCurrentSymbol();
		    if ( (!fsys.get(sym)) && (sym != SetSym.IDENT) ) {
			    error(BugBox.BAD_SYMB);
			    do {
				    lexicalA.getNextSymbol();
			    } while (!fsys.get(lexicalA.getCurrentSymbol()));
		    }
		    if (lexicalA.getCurrentSymbol() == SetSym.IDENT) {
			   
			    i = symbol.position(lexicalA.getCurrentIdentifier(),         
			                        level);                                  
			    if (i <= 0) {
				    error(BugBox.UNDECLARED_ID);
			    } else if (!((symbol.getIdentifier(i).GetType()).IsInstanceOfType(new Variable(null,0,0)))) {
				    /* prirazeni do jineho ident. nez promenna */
				    error(BugBox.WRONG_BECOMES);
				    i = 0;
			    }
			    lexicalA.getNextSymbol();
			    if (lexicalA.getCurrentSymbol() == SetSym.BECOMES) {
				    lexicalA.getNextSymbol();
			    } else {
				    error(BugBox.BECOMES);
			    }
			    expression(fsys,tx,level);
			    sym = lexicalA.getCurrentSymbol();
			    if (i != 0) {
				    code.gen(Fct.STO,
				             level-((Variable) symbol.getIdentifier(i)).getLevel(),
				             ((Variable) symbol.getIdentifier(i)).getAddress());
			    }
		    } else if (sym == SetSym.CALLSYM) {
			    /* nalezeno volani podprogramu */
			    lexicalA.getNextSymbol();
			    if (lexicalA.getCurrentSymbol() != SetSym.IDENT) {
				    error(BugBox.ID_PAS_CALL);
			    } else {
				    /* Doplen parametr funkce na nasledujici radce */        
				    i = symbol.position(lexicalA.getCurrentIdentifier(),     
				                        level);                              
				    if (i == 0) {
						    error(BugBox.UNDECLARED_ID);
				    } else {
					    if ((symbol.getIdentifier(i).GetType()).IsInstanceOfType(new Procedure(null,0))) {
						    code.gen(Fct.CAL,
						             level- ((Procedure)
						               symbol.getIdentifier(i)).getLevel(),
						             ((Procedure)
						               symbol.getIdentifier(i)).getAddress());
					    } else {
						    error(BugBox.INVALID_CALL);
					    }
				    }
			    lexicalA.getNextSymbol();
			    }
		    } else if (lexicalA.getCurrentSymbol() == SetSym.IFSYM) {
			    /* podmineny prikaz */
			    lexicalA.getNextSymbol();
			    pom = new SetSym();
			    pom = pom.or(fsys);
			    pom.set(SetSym.THENSYM);
			    pom.set(SetSym.DOSYM);
			    condition(pom,tx,level);
			    if (lexicalA.getCurrentSymbol() == SetSym.THENSYM) {
				    lexicalA.getNextSymbol();
			    } else {
				    error(BugBox.THEN_REQ);
			    }
			    cx1 = code.getAddressCounter();
			    code.gen(Fct.JPC, 0, 0);
			    statement(fsys, tx, level);
			    code.setAddressAt(cx1, code.getAddressCounter());
		    } else if (lexicalA.getCurrentSymbol() == SetSym.BEGINSYM) {
			    /* zacina novy blok */
			    lexicalA.getNextSymbol();
			    pom = new SetSym();
			    pom = pom.or(fsys);
			    pom.set(SetSym.SEMICOLON);
			    pom.set(SetSym.ENDSYM);
			    statement(pom, tx, level);
			    sym = lexicalA.getCurrentSymbol();
                while ((bs.statBegSys.get(sym)) || (sym == SetSym.SEMICOLON))
                {
				    if (sym == SetSym.SEMICOLON) {
					    lexicalA.getNextSymbol();
					    sym = lexicalA.getCurrentSymbol();
				    } else {
					    error(BugBox.BAD_SYMB);
				    }
				    pom = pom.or(fsys);
				    pom.set(SetSym.SEMICOLON);
				    pom.set(SetSym.ENDSYM);
				    statement(pom, tx, level);
				    sym = lexicalA.getCurrentSymbol();
			    }
			    if (sym == SetSym.ENDSYM) {
				    lexicalA.getNextSymbol();
			    } else {
				    error(BugBox.END_BLOCK_REQ);
			    }
		    } else if (sym == SetSym.WHILESYM) {
			    /* zacina cyklus while */
			    cx1 = code.getAddressCounter();
			    lexicalA.getNextSymbol();
			    pom = new SetSym();
			    pom = pom.or(fsys);
			    pom.set(SetSym.DOSYM);
			    condition(pom, tx, level);
			    cx2 = code.getAddressCounter();
			    code.gen(Fct.JPC, 0, 0);
			    if (lexicalA.getCurrentSymbol() == SetSym.DOSYM) {
				    lexicalA.getNextSymbol();
			    } else {
				    error(BugBox.DO_REQ);
			    }
			    statement(fsys, tx, level);
			    code.gen(Fct.JMP, 0, cx1);
			    code.setAddressAt(cx2, code.getAddressCounter());
		    }
		    test(fsys, pom, 19); /* test koreknost ukonceni statementu */
	    }
       
        /// <summary>
        /// Testuje zda nacteny symbol je v mnozine symbolu 's1'
        /// Pokud neni generuje chybu a nacita opakovane ze vstupu
        /// dokud neni nacten pristupny symbol v mnozina 's1' a 's2'
        /// </summary>
        /// <param name="s1">predpokladana mnozina znaku</param>
        /// <param name="s2">akceptovatelna mnozina znaku</param>
        /// <param name="n">cislo chyby</param>
        private void test(SetSym s1, SetSym s2, int n)
        {
            // pomocna referencni promenna na mnozinu znaku.
            SetSym pom = new SetSym();

            if (!s1.get(lexicalA.getCurrentSymbol()))
            {
                error(n);
                pom = pom.or(s1);
                pom = pom.or(s2);
                while (!pom.get(lexicalA.getCurrentSymbol()))
                    lexicalA.getNextSymbol();
            }
        }

    }
}
