﻿using System;
using System.Collections.Generic;
using System.Text;

namespace pl0
{
    /// <summary>
    /// Spravuje chyby prekladace
    /// </summary>
    public class BugBox
    {
	    private int errors = 0;
	    
	    private int maxErrors = 30;

        public const int BAD_BECOMES = 1;
        public const int NUM_REQ = 2;
        public const int EQUALS_REQ = 3;
        public const int ID_REQ = 4;
        public const int SEMI_COMO_REQ = 5;
        public const int BAD_PAST_PROC = 6;
	    public const int STATEMENT_REQ = 7;
	    public const int BAD_PAST_STATEMENT = 8;
	    public const int DOT_REQ = 9;
	    public const int BAD_SYMB = 10;
	    public const int UNDECLARED_ID = 11;
	    public const int WRONG_BECOMES = 12;
	    public const int BECOMES = 13;
	    public const int ID_PAS_CALL = 14;
	    public const int INVALID_CALL = 15;
	    public const int THEN_REQ = 16;
	    public const int END_BLOCK_REQ = 17;
	    public const int DO_REQ = 18;
	    public const int INVALID_SYMB = 19;
	    public const int RELATION_REQ = 20;
	    public const int INVALID_PROC_USE = 21;
	    public const int RPARENT_REQ = 22;
	    public const int INVALID_FACTOR_END = 23;
	    public const int INVALID_EXPRESS_BEG = 24;
	    public const int TOO_BIG_NUMBER = 30;
	    public const int TOO_BIG_LEVEL = 32;

        public void error(int errorCode)
        {
            errors++;
            if (errors > maxErrors)
                throw new ErrorsException();
        }

        /// <summary>
        /// Celkovy pocet chyb v programu
        /// </summary>
        /// <returns>celkovy pocet chyb v programu</returns>
        public int getErrorsCount()
        {
            return errors;
        }

        public void setMaxErrors(int max)
        {
            maxErrors = max;
            if (errors > maxErrors)
                throw new ErrorsException();
        }
       
        public int getMaxErrors()
        {
            return maxErrors;
        }

        /// <summary>
        /// Vraci chybovy text
        /// </summary>
        /// <param name="errorCode">cislo chyby</param>
        /// <returns>text chyby</returns>
        public static String getErrorText(int errorCode)
        {
            switch (errorCode)
            {
                case BAD_BECOMES:
                    return ("Pouzito \"=\" misto \":=\"");
                case NUM_REQ:
                    return ("Za \"=\" musi nasledovat cislo");
                case EQUALS_REQ:
                    return ("Za identifikatorem ma nasledovat \"=\"");
                case ID_REQ:
                    return ("Za \"const\", \"var\", \"procedure\" " +
                           "musi nasledovat identifikator");
                case SEMI_COMO_REQ:
                    return ("Chybi strednik nebo carka");
                case BAD_PAST_PROC:
                    return ("Nespravny symbol po deklaraci procedury");
                case STATEMENT_REQ:
                    return ("Je ocekavan prikaz");
                case BAD_PAST_STATEMENT:
                    return ("Neocekavany symbol za prikazovou casti bloku");
                case DOT_REQ:
                    return ("Ocekavam tecku");
                case BAD_SYMB:
                    return ("Nespravny symbol v prikazu");
                case UNDECLARED_ID:
                    return ("Nedeklarovany identifikator");
                case WRONG_BECOMES:
                    return ("Prirazeni konstante a procedure neni dovoleno");
                case BECOMES:
                    return ("Operator prirazeni je \":=\"");
                case ID_PAS_CALL:
                    return ("Za \"call\" musi nasledovat identifikator");
                case INVALID_CALL:
                    return ("Volani konstanty nebo promenne neni dovoleno");
                case THEN_REQ:
                    return ("Ocekavano \"then\"");
                case END_BLOCK_REQ:
                    return ("Ocekavano \"end\" nebo \";\"");
                case DO_REQ:
                    return ("Ocekavano \"do\"");
                case INVALID_SYMB:
                    return ("Nespravne pouzity symbol za prikazem");
                case RELATION_REQ:
                    return ("Ocekavam relaci");
                case INVALID_PROC_USE:
                    return ("Jmeno procedury nelze pouzit ve vyrazu");
                case RPARENT_REQ:
                    return ("Chybi uzaviraci zavorka");
                case INVALID_FACTOR_END:
                    return ("Faktor nesmi koncit timto symbolem");
                case INVALID_EXPRESS_BEG:
                    return ("Vyraz nesmi zacinat timto symbolem");
                case TOO_BIG_NUMBER:
                    return ("Prilis velke cislo");
                case TOO_BIG_LEVEL:
                    return ("Prilis velke zanoreni");
            }
            return ("Neznama chyba");
        }
    }
}
