﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace pl0
{
    /// <summary>
    /// Tabulka generovanych instrukci
    /// </summary>
    public class CodeGen
    {
        
	    private static int MAX_INSTRUCTION = 1000;
	    
	    private static String[] MNEMONIC = {"LIT  ","OPR  ","LOD  ","STO  ","CAL  ",
	                                       "RET  ","INT  ","JMP  ","JMC  "};
        
        private GenInstruction[] genCode = new GenInstruction[MAX_INSTRUCTION];
        
        private int addressCounter = 0;

        
        public void setAddressAt(int where, int what)
        {
            genCode[where].operand = what;
        }
        
        public int getAdress(int where)
        {
            if (genCode[where] != null)
            {
                return genCode[where].operand;
            }
            else
            {
                return 0;
            }
        }
        
        public int getAddressCounter()
        {
            return addressCounter;
        }

        /// <summary>
        /// Pridani instrukce do tabulky instrukci
        /// </summary>
        /// <param name="x">kod instrukce</param>
        /// <param name="y">uroven zanoreni</param>
        /// <param name="z">adresni cast v instrukci</param>
        public void gen(int x, int y, int z) 
        {
		    if (addressCounter >= MAX_INSTRUCTION) {
			    Console.Out.WriteLine("Program je prilis dlouhy!");
                Environment.Exit(1);
		    }
		    genCode[addressCounter++] = new GenInstruction(x, y, z); 
	    }

        /// <summary>
        /// Vypis generovaneho kodu
        /// </summary>
        /// <param name="path">Soubor kam se bude vypisovat</param>
        public void list(String path) 
        {
            try
            {
                using (StreamWriter sw = new StreamWriter(path))
                {
                    sw.WriteLine("Generovany kod:");
                    sw.WriteLine("**************");
                    for (int i = 0; i <= (addressCounter-1); i++) {
                        sw.WriteLine(i +"  "+ MNEMONIC[genCode[i].functionCode] + genCode[i].level +"  "+ genCode[i].operand);
			        }
                }
            }
            catch (IOException e)
            {
                Console.Out.WriteLine("Chyba pri zapisovani do souboru" + path);
                Console.Out.WriteLine(e.StackTrace);
            }
	    }

       
       /// <summary>
       /// Vypis casti generovaneho kodu
       /// </summary>
       /// <param name="from">index od kud se bude vypisovat</param>
        public void listCode(int from, int to)
        {
		    for (int i = from; i <= (addressCounter-1); i++) {
			    
                Console.Out.WriteLine(i +"  "+ MNEMONIC[genCode[i].functionCode] + genCode[i].level +"  "+ genCode[i].operand);
		    }
	    }

        public GenInstruction getInstruction(int i)
        {
            return genCode[i];
        }

    }
}
