﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.IO;

namespace pl0
{
    /// <summary>
    /// Realizuje lexikalni analizu zpracovavaneho souboru
    /// </summary>
    public class LexicalA
    {
        /// <summary>
        /// Naposledy nacteny simbol
        /// </summary>
        private int currentSymbol;
        /// <summary>
        /// Naposledy nactene cislo
        /// </summary>
        private int currentNumber;
        
        private bool lastNumberTooBig;                                    
        
        private SourceReader source;
        /// <summary>
        /// Naposledy nacteny identifikator
        /// </summary>
        private StringBuilder currentIdentifier = new StringBuilder();
        /// <summary>
        /// Naposledy nacteny znak
        /// </summary>
        private Char currentChar = ' ';
        /// <summary>
        /// Tabulka pro jednomistne simboly
        /// </summary>
	    private Hashtable singleSymbols = new Hashtable();
	    /// <summary>
	    /// Tabulka pro vycemistne simboly
	    /// </summary>
	    private Hashtable keyWords = new Hashtable();

        public int getCurrentSymbol()
        {
		    return currentSymbol;
	    }
	    
        public int getCurrentNumber()
        {
		    return currentNumber;
	    }

        public bool isLastNumberTooBig()
        {                                
		    return lastNumberTooBig;                                         
	    }                                                                    

        public char getCurrentChar()
        {
		    return currentChar;
	    }	

        public String getCurrentIdentifier()
        {
            return currentIdentifier.ToString();
	    }
        
        /// <summary>
        /// Vypise misto a popis chyby
        /// </summary>
        /// <param name="errorType">cislo chyby</param>
        public void writeError(int errorType)
        {
		    for (int i = 0; i < source.currentIndex-2; i++)
			    Console.Out.Write(" ");
		    Console.Out.WriteLine("^ " + errorType + " - " + BugBox.getErrorText(errorType));
	    }
        /// <summary>
        /// Konstruktor
        /// </summary>
        /// <param name="path">soubor se zpracovavanym zdrojovym textem</param>
	    public LexicalA(String path) /*throw IOException*/ {
		    source = new SourceReader(path);

		    singleSymbols.Add(('+'), (SetSym.PLUS));
		    singleSymbols.Add(('-'), (SetSym.MINUS));
            singleSymbols.Add(('*'), (SetSym.TIMES));
            singleSymbols.Add(('/'), (SetSym.SLASH));
            singleSymbols.Add(('%'), (SetSym.MODULO));
            singleSymbols.Add(('('), (SetSym.LPAREN));
            singleSymbols.Add((')'), (SetSym.RPAREN));
            singleSymbols.Add(('='), (SetSym.EQL));
            singleSymbols.Add((','), (SetSym.COMMA));
            singleSymbols.Add(('.'), (SetSym.PERIOD));
            singleSymbols.Add(('#'), (SetSym.NEQ));
            singleSymbols.Add(('<'), (SetSym.LSS));
            singleSymbols.Add(('>'), (SetSym.GTR));
            singleSymbols.Add((';'), (SetSym.SEMICOLON));

            keyWords.Add(("begin"), (SetSym.BEGINSYM));
            keyWords.Add(("call"), (SetSym.CALLSYM));
            keyWords.Add(("const"), (SetSym.CONSTSYM));
            keyWords.Add(("do"), (SetSym.DOSYM));
            keyWords.Add(("end"), (SetSym.ENDSYM));
            keyWords.Add(("if"), (SetSym.IFSYM));
            keyWords.Add(("odd"), (SetSym.ODDSYM));
            keyWords.Add(("procedure"), (SetSym.PROCSYM));
            keyWords.Add(("then"), (SetSym.THENSYM));
            keyWords.Add(("var"), (SetSym.VARSYM));
            keyWords.Add(("while"), (SetSym.WHILESYM));

		    getNextSymbol();  // automaticke nacteni prvniho znaku
	    }
        
        public void zeroNumber()
        {
            currentNumber = 0;
        }

        /// <summary>
        /// Nacteni dalsiho simbolu
        /// </summary>
        public void getNextSymbol()
        {
            while (currentChar <= ' ')
                currentChar = source.getChar(); /* Nacteni jednoho znaku */
            // nacteni identifikatoru do currentIdentifier
            if (Char.IsLetter(currentChar))
            {
                /* identifier or reserved word */
                currentIdentifier = new StringBuilder();
                do
                {
                    currentIdentifier.Append((currentChar));
                    currentChar = source.getChar();  /* Nacteni jednoho znaku */
                } while (Char.IsLetterOrDigit(currentChar));
                if (keyWords.ContainsKey((string)currentIdentifier.ToString()) == true)
                {
                    currentSymbol = (int)keyWords[(string)currentIdentifier.ToString()];
                }
                else
                {
                    currentSymbol = SetSym.IDENT;
                }
            }
            else if (Char.IsDigit(currentChar))
            {
                /* number */
                currentSymbol = SetSym.NUMBER;
                currentNumber = 0;
                long cislo = 0;                                              
               
                lastNumberTooBig = false;                                    
                do
                {  /* get value */
                    cislo = (10 * cislo) + (currentChar - '0');
                    currentChar = source.getChar();
                    /* Kontrola velikosti zadaneho cisla */
                    
                    if ((cislo > int.MaxValue) ||                      
                         (cislo < int.MinValue))                       
                        lastNumberTooBig = true;                             
                } while (Char.IsDigit(currentChar));
                if (!lastNumberTooBig)                                       
                    currentNumber = (int)cislo;                             
            }
            else if (currentChar == ':')
            {
                /* muze jit o prirazeni */
                currentChar = source.getChar();
                if (currentChar == '=')
                {
                    currentSymbol = SetSym.BECOMES;
                    currentChar = source.getChar();
                }
                else
                {
                    currentSymbol = SetSym.NULL;
                }
            }
            else if (currentChar == '<')
            {
                currentChar = source.getChar();
                if (currentChar == '=')
                {
                    currentSymbol = SetSym.LEQ;  /* mensi nebo rovno */
                    currentChar = source.getChar();
                }
                else
                {
                    currentSymbol = SetSym.LSS;  /* mensi nez */
                }
            }
            else if (currentChar == '>')
            {
                currentChar = source.getChar();
                if (currentChar == '=')
                {
                    currentSymbol = SetSym.GEQ;  /* vetsi nebo rovno */
                    currentChar = source.getChar();
                }
                else
                {
                    currentSymbol = SetSym.GTR;  /* vetsi nez */
                }
            }
            else
            {
                if (singleSymbols.ContainsKey((char)(currentChar)))
                {
                    currentSymbol = (int)singleSymbols[(char)(currentChar)];
                    currentChar = source.getChar();
                }
                else
                {
                    currentSymbol = SetSym.NULL;
                    currentChar = source.getChar();
                }
            }
        }

      /// <summary>
      /// Slouzi k praci se souborem
      /// </summary>
	  private class SourceReader {

		private StreamReader sr = null;
		/* Aktualni radka programu */
        public String currentLine = "";
		/* Aktualni offset vzhledem k zacatku radky */
        public int currentIndex = 0;

		public SourceReader(String fileName) /*throws IOException*/ 
        {
			try {
				sr = new StreamReader(fileName);
			} catch (FileNotFoundException e) {
				Console.Out.WriteLine("Soubor " + fileName + " nenalezen.");
				throw e;
			}
		}
        
        /// <summary>
        /// Nacete a vrati dalsi znak z souboru
        /// </summary>
        /// <returns>dalsi znak z souboru</returns>
		public char getChar() 
        {
			if (currentIndex == currentLine.Length) {
				try {
					if ((currentLine = sr.ReadLine()) == null) {
						Console.Out.WriteLine("Program neni kompletni nebo " +
						                   "chybi tecka za koncovym end.");
						throw new ErrorsException();
					} else {
						currentLine = (currentLine + " ");
						currentIndex = 0;
						Console.Out.WriteLine(currentLine);
						return currentLine[currentIndex++];
					}
				} catch (IOException e) {
					Console.Out.WriteLine("Chyba pri cteni ze souboru.");
					Console.Out.WriteLine(e.StackTrace);
					return (char)0;
				}
			} else {
                return currentLine[currentIndex++];
			}
		}

		
		public void finalize() /*throws Throwable*/ 
        {
			if (sr != null)
				sr.Close();
			
		}
	  }  // end of SourceReader
    }
}
