﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace pl0
{
    /// <summary>
    /// Interpret jazyka PL0
    /// </summary>
    public class PL0Interpret
    {
	    private const int STACK_SIZE = 500;
	    
	    private CodeGen code;
	    
	    private SymbolT symbol;
	    /// <summary>
	    /// Zasobnik
	    /// </summary>
	    private int[] stack = new int[STACK_SIZE];
	    /// <summary>
	    /// Aktualni uroven zanoreni
	    /// </summary>
	    private int currentLevel;

        public PL0Interpret(SymbolT sym, CodeGen code)
        {
            this.code = code;
            this.symbol = sym;
        }
       
        /// <summary>
        /// Metoda pro zjisteni pocatecni adresy na urovni, ktera je
        /// o l vyssi nez soucasna uroven, ve ktere je prave vykonan program.
        /// </summary>
        /// <param name="l">rozdil pozadovane a aktualni urovne</param>
        /// <returns>adresa pocatecni urovne</returns>
	    private int baze(int l) 
        {
		    int baseLevel = currentLevel;
		    while (l > 0) {
			    baseLevel = stack[baseLevel];
			    l--;
		    }
		    return baseLevel;
	    }

        /// <summary>
        /// Spusteni interpretu
        /// </summary>
        public void run() 
        {
		    int p;  // Citac instrukci
		    int t;  // Ukazatel na vrchol zasobniku
		    GenInstruction i;

		    Console.Out.WriteLine("START PL/0\n");
		    t = p = stack[1] = stack[2] = stack[3] = 0;  currentLevel=1;
		    do {
			    i = code.getInstruction(p++);
			    switch (i.functionCode) {
			    case Fct.LIT:
				    stack[++t] = i.operand;
				    break;
			    case Fct.OPR:
				    switch (i.operand) {
				    case OOperation.NEG:
					    stack[t] =- stack[t];
					    break;
				    case OOperation.ADD:
					    t--;
					    stack[t] += stack[t+1];
					    break;
				    case OOperation.SUB:
					    t--;
					    stack[t] -= stack[t+1];
					    break;
				    case OOperation.MUL:
					    t--;
					    stack[t] *= stack[t+1];
					    break;
				    case OOperation.DI:
					    t--;
					    stack[t] /= stack[t+1];
					    break;
				    case OOperation.MOD:
					    t--;
					    stack[t] = stack[t] % stack[t+1];
					    break;
				    case OOperation.ODD:
					    stack[t] = stack[t] % 2;
					    break;
				    case OOperation.EQ:
					    t--;
					    stack[t] = (stack[t] == stack[t+1]) ? 1 : 0;
					    break;
				    case OOperation.NE:
					    t--;
					    stack[t] = (stack[t] != stack[t+1]) ? 1 : 0;
					    break;
				    case OOperation.LT:
					    t--;
					    stack[t] = (stack[t] < stack[t+1]) ? 1 : 0;
					    break;
				    case OOperation.GE:
					    t--;
					    stack[t] = (stack[t] >= stack[t+1]) ? 1 : 0;
					    break;
				    case OOperation.GT:
					    t--;
					    stack[t] = (stack[t] > stack[t+1]) ? 1 : 0;
					    break;
				    case OOperation.LE:
					    t--;
					    stack[t] = (stack[t] <= stack[t+1]) ? 1 : 0;
					    break;
				    }
				    break;
			    case Fct.LOD:
				    t++;
				    stack[t] = stack[baze(i.level)+i.operand];
				    break;
			    case Fct.STO:
				    stack[baze(i.level)+i.operand] = stack[t];
				    Console.Out.WriteLine(stack[t--]);
				    break;
			    case Fct.CAL:
				    stack[t+1] = baze(i.level);
				    stack[t+2] = currentLevel;
				    stack[t+3] = p;
				    currentLevel = t+1;
				    p = i.operand;
				    break;
			    case Fct.RET:
				    t = currentLevel-1;
				    p = stack[t+3];
				    currentLevel = stack[t+2];
				    break;
			    case Fct.ING:
				    t += i.operand;
				    break;
			    case Fct.JMP:
				    p = i.operand;
				    break;
			    case Fct.JPC:
				    if (stack[t] == 0)
					    p = i.operand;
				    t--;
				    break;
			    }
		    } while (p != 0);

		    try {
			    code.list("LIST.TAB");
			    symbol.list("SYMBOL.TAB");
		    } catch (IOException e) {
			    Console.Out.WriteLine("Chyba pri zapisu do pomocnych souboru\n");
			    Console.Out.WriteLine(e.StackTrace);
		    }
		    Console.Out.WriteLine(" END PL/0\n");
	    }


    }
}
