﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace pl0
{
    /// <summary>
    /// Symboly jazyka PL0
    /// </summary>
    public class SetSym
    {
        public  const int NULL = 0;
	    public  const int IDENT = 1;
        public  const int NUMBER = 2;
        public  const int PLUS = 3;
        public  const int MINUS = 4;
        public  const int TIMES = 5;
        public  const int SLASH = 6;
        public  const int MODULO = 7;
        public  const int ODDSYM = 8;
        public  const int EQL = 9;
        public  const int NEQ = 10;
        public  const int LSS = 11;
        public  const int LEQ = 12;
        public  const int GTR = 13;
        public  const int GEQ = 14;
        public  const int LPAREN = 15;
        public  const int RPAREN = 16;
        public  const int COMMA = 17;
        public  const int SEMICOLON = 18;
        public  const int PERIOD = 19;
        public  const int BECOMES = 20;
        public  const int BEGINSYM = 21;
        public  const int ENDSYM = 22;
        public  const int IFSYM = 23;
        public  const int THENSYM = 24;
        public  const int WHILESYM = 25;
        public  const int DOSYM = 26;
        public  const int CALLSYM = 27;
        public  const int CONSTSYM = 28;
        public  const int VARSYM = 29;
        public  const int PROCSYM = 30;
        public  const int SYMBOLS = 31;

        /// <summary>
        /// Mnozina simbolu
        /// </summary>
        private BitArray languageSet = new BitArray(SYMBOLS);

        /// <summary>
        /// vlozi simbol do mnoziny (nastavi na danem indexu true)
        /// </summary>
        /// <param name="bit">vkladany simbol</param>
	    public void set(int bit) 
        {
		    if ( (bit >= 0) && (bit <= SYMBOLS-1) )
			    languageSet.Set(bit, true);	
	    }
        
        /// <summary>
        /// vyjme simbol do mnoziny (nastavi na danem indexu false)
        /// </summary>
        /// <param name="bit">simbol</param>
	    public void clear(int bit) 
        {
		    if ( (bit >= 0) && (bit <= SYMBOLS-1) )
			    languageSet.Set(bit, false);
	    }

        /// <summary>
        /// Urci yda se simbol nachazi v mnozine
        /// </summary>
        /// <param name="bit">simbol</param>
        /// <returns>TRUE pokud se simbol nachazi v mnozine</returns>
        public bool get(int bit) 
        {
		    if ( (bit >= 0) && (bit <= SYMBOLS-1) ) {
			    return languageSet.Get(bit);
		    } else {
			    return false;
		    }
	    }

        /// <summary>
        /// Vytvori novou mnozinu jako prunik s mnonou sset
        /// </summary>
        /// <param name="sset">mnozina</param>
        /// <returns>nove vytvorena mnozina</returns>
        public SetSym and(SetSym sset)
        {
		    SetSym retvalue = new SetSym();
		    retvalue.languageSet.Or(languageSet);
		    retvalue.languageSet.And(sset.languageSet);
		    return retvalue;
	    }

        //// <summary>
        /// Vytvori novou mnozinu jako sjednoceni s mnonou sset
        /// </summary>
        /// <param name="sset">mnozina</param>
        /// <returns>nove vytvorena mnozina</returns>
        public SetSym or(SetSym sset)
        {
		    SetSym retvalue = new SetSym();
		    retvalue.languageSet.Or(languageSet);
		    retvalue.languageSet.Or(sset.languageSet);
		    return retvalue;
	    }

	    /// <summary>
	    /// V mnozine nebude zadny simbol
	    /// </summary>
        public void zero()
        {
            languageSet.SetAll(false);
	    }
    }
}
