﻿// oprava v setAddr, puvodne kod spatne rozpoznaval, ze procedura je Locatable

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace pl0
{
    /// <summary>
    /// Tabulka symbolu
    /// </summary>
    public class SymbolT
    {
	    private const int TX_MAX = 5000;
        
        private Identifier[] table = new Identifier[TX_MAX];
        
        private int tableIndex = 0;
        
        /// <summary>
        /// Vyhleda symbol v tabulce symbolu
        /// </summary>
        /// <param name="id">jmeno symbolu</param>
        /// <param name="level">uroven zanoreni</param>
        /// <returns>0: symbol nenalezen, >0 : adresa symbolu</returns>
        public int position(String id, int level)
        {  
            for (int i = tableIndex; i > 0; i--)
            {                          
                /* Kontrola urovne zanoreni */
                if (table[i].level > level)                                  
                    continue;  // To je moc vnorene, to nevidime             
                if (table[i].level < level)                                  
                    level = table[i].level;  // Vynorujeme se                
                /* Porovnani symbolu */
                if (table[i].name.CompareTo(id) == 0)                        
                    return (i);  // Nalezeno                                 
            }                                                                
            return (0);                                                      
        }

        /// <summary>
        /// Vlozeni konstanty do tabulky symbolu.
        /// </summary>
        /// <param name="identifier">jmeno konstanty</param>
        /// <param name="level">uroven zanoreni</param>
        /// <param name="value">hodnota konstanty</param>
        public void enterConstant(String identifier, int level, int value) 
        {                            
	        /* Doplnena kontrola preteceni tabulky symbolu */                
		    if (tableIndex >= (TX_MAX-1)) {                                  
			    Console.Out.WriteLine("Pocet symbolu je prilis velky!");        
			    //System.exit(1);                                              
                Environment.Exit(1);
		    }                                                               
		    tableIndex++;
		    table[tableIndex] = new Constant(identifier, level, value);
	    }
        
        /// <summary>
        /// Vlozeni promenne do tabulky symbolu.
        /// </summary>
        /// <param name="identifier">jmeno promenne</param>
        /// <param name="address">adresa</param>
        /// <param name="level">uroven zanoreni</param>
        public void enterVariable(String identifier, int address, int level)
        {
	        /* Doplnena kontrola preteceni tabulky symbolu */                
		    if (tableIndex >= (TX_MAX-1)) {                                  
			    Console.Out.WriteLine("Pocet symbolu je prilis velky!");       
                Environment.Exit(1);
		    }                                                                
		    tableIndex++;
		    table[tableIndex] = new Variable(identifier, address, level);
	    }
        
        /// <summary>
        /// Vlozeni procedury do tabulky symbolu 
        /// </summary>
        /// <param name="identifier">jmeno procedury</param>
        /// <param name="level">uroven zanoreni</param>
        public void enterProcedure(String identifier, int level)
        {
	        /* Doplnena kontrola preteceni tabulky symbolu */               
		    if (tableIndex >= (TX_MAX-1)) {                                  
                Console.Out.WriteLine("Pocet symbolu je prilis velky!");
                Environment.Exit(1);
		    }                                                               
		    tableIndex++;
		    table[tableIndex] = new Procedure(identifier, level);
	    }
        
        /// <summary>
        /// Nastaveni adresy symbolu.
        /// </summary>
        /// <param name="tx">offset symbolu v tabulce</param>
        /// <param name="cx">nova hodnota adresy</param>
        public void setAddr(int tx, int cx)
        {
		    if (table[tx] == null) {
			    table[tx] = new Variable("", cx, 0);
            }
            else if (table[tx].GetType().IsSubclassOf(typeof(Locatable))) // PL opraveno
            {
			    ((Locatable) table[tx]).setAddress(cx);
		    }
	    }
       
        /// <summary>
        /// Ziskani prislusneho identifikatoru z tabulky identifikatoru 
        /// </summary>
        /// <param name="i">offset symbolu</param>
        /// <returns>symbol</returns>
        public Identifier getIdentifier(int i)
        {
		    return table[i];
	    }

        public int getTableIndex()
        {
		    return tableIndex;
	    }

	    /// <summary>
	    /// Vypis tabulky symbolu do souboru
	    /// </summary>
	    /// <param name="path">jmeno souboru</param>
	    public void list(String path)/* throws IOException*/ {
		    try {
			    
                using(StreamWriter sw = new StreamWriter(path)){
                    sw.WriteLine("Tabulka symbolu:");
                    sw.WriteLine("***************");
                    for (int i = 1; i <= tableIndex; i++) {
				        sw.Write(i + " ");
				        sw.WriteLine(table[i].ToString());
			        }
                }
		    } catch (IOException e) {
			    Console.Out.WriteLine("File " + path + " couldn't be created.");
			    throw e;
		    }
	    }
    }
}
